module abstract_vector

  implicit none

  private
  public :: vector_class

  ! VECTOR_CLASS is an abstract type for vectors.

  ! VECTOR_CLASS defines the following method:
  !   assign(this,v)      : (deferred) Copy v to this.
  !   negate(this)        : (deferred) negate the vector elements.
  !   subtract(this,v)    : (deferred) subtract v from this.
  !   add(this,v)         : (deferred) add v to this
  !   multiply_by(this,s) : (deferred) multiply elements of this by s.
  !   saxpy(this,s,v)     : (deferred) this = this + s*v

  type, abstract :: vector_class
    private
  contains
    private
    procedure(op_assign_v_v), public, deferred :: assign
    procedure(op_v),          public, deferred :: negate
    procedure(op_v_v),        public, deferred :: subtract
    procedure(op_v_v),        public, deferred :: add
    procedure(op_v_s),        public, deferred :: multiply_by
    procedure(op_v_s_v),      public, deferred :: saxpy
  end type vector_class

  abstract interface
    subroutine op_assign_v_v(this,v)
      import vector_class
      class(vector_class), intent(inout) :: this
      class(vector_class), intent(in)    :: v
    end subroutine
    subroutine op_v (this)
      import :: vector_class
      class(vector_class),intent(inout)  :: this
    end subroutine
    subroutine op_v_v (this,v)
      import :: vector_class
      class(vector_class),intent(inout)  :: this
      class(vector_class),intent(in)     :: v
    end subroutine
    subroutine op_v_s (this,s)
      import :: vector_class
      class(vector_class),intent(inout)  :: this
      real, intent(in)                          :: s
    end subroutine
    subroutine op_v_s_v (this,s,v)
      import :: vector_class
      class(vector_class),intent(inout)  :: this
      real, intent(in)                          :: s
      class(vector_class),intent(in)     :: v
    end subroutine
  end interface
end module abstract_vector
