module concrete_inner_product
  use abstract_vector
  use abstract_inner_product
  use concrete_vector
  use concrete_gradient
  implicit none

  ! TRIVIAL_INNER_PRODUCT_TYPE provides a simple concrete example of the
  ! abstract type defined by INNER_PRODUCT_CLASS. Other definitions are
  ! possible. The abstract conjugate gradients routine will work with all
  ! of them!

  type, extends(inner_product_class) :: trivial_inner_product_type
    real :: scalp(100)  
    real :: one_over_scalp(100)  
  contains
      procedure :: dot_v_v           => my_dot_v_v
      procedure :: dot_g_g           => my_dot_g_g
      procedure :: D_times_v         => my_D_times_v
      procedure :: D_times_g         => my_D_times_g
      procedure :: D_inverse_times_v => my_D_inverse_times_v
      procedure :: D_inverse_times_g => my_D_inverse_times_g
  end type trivial_inner_product_type

contains

  function my_dot_v_v (this,a,b)
    implicit none
    class(trivial_inner_product_type), intent(in) :: this
    class(vector_class),               intent(in) :: a, b
    real :: my_dot_v_v

    select type (a)
    type is (trivial_vector_type)
      select type (b)
      type is (trivial_vector_type)
        my_dot_v_v = dot_product(a%elements,b%elements)
      end select
    end select
  end function my_dot_v_v

  function my_dot_g_g (this,a,b)
    implicit none
    class(trivial_inner_product_type), intent(in) :: this
    class(vector_class),               intent(in) :: a, b
    real :: my_dot_g_g

    select type (a)
    type is (trivial_gradient_type)
      select type (b)
      type is (trivial_gradient_type)
        my_dot_g_g = dot_product(a%elements,b%elements)
      end select
    end select
  end function my_dot_g_g

  subroutine my_D_times_v (this,a)
    implicit none
    class(trivial_inner_product_type), intent(in)    :: this
    class(vector_class),               intent(inout) :: a

    select type (a)
    type is (trivial_vector_type)
      a%elements(:) = a%elements(:)*this%scalp(:)
    end select
  end subroutine my_D_times_v

  subroutine my_D_times_g (this,a)
    implicit none
    class(trivial_inner_product_type), intent(in)    :: this
    class(vector_class),               intent(inout) :: a

    select type (a)
    type is (trivial_gradient_type)
      a%elements(:) = a%elements(:)*this%scalp(:)
    end select
  end subroutine my_D_times_g

  subroutine my_D_inverse_times_v (this,a)
    implicit none
    class(trivial_inner_product_type), intent(in)    :: this
    class(vector_class),               intent(inout) :: a

    select type (a)
    type is (trivial_vector_type)
      a%elements(:) = a%elements(:)*this%one_over_scalp(:)
    end select
  end subroutine my_D_inverse_times_v

  subroutine my_D_inverse_times_g (this,a)
    implicit none
    class(trivial_inner_product_type), intent(in)    :: this
    class(vector_class),               intent(inout) :: a

    select type (a)
    type is (trivial_gradient_type)
      a%elements(:) = a%elements(:)*this%one_over_scalp(:)
    end select
  end subroutine my_D_inverse_times_g

end module concrete_inner_product
