module concrete_simulator
  use abstract_vector
  use abstract_gradient
  use abstract_inner_product
  use abstract_simulator
  use concrete_vector
  use concrete_gradient
  use concrete_inner_product
  implicit none

  ! TRIVIAL_SIMULATOR_TYPE provides a simple concrete example of the abstract
  ! type defined by SIMULATOR_CLASS. Other definitions are possible. The
  ! abstract conjugate gradients routine will work with all of them!

  type, extends(simulator_class) :: trivial_simulator_type
    private
  contains
    procedure :: eval => my_eval
  end type trivial_simulator_type

contains

  subroutine my_eval(this,x,g,ip)
    implicit none
    class(trivial_simulator_type),     intent(in)    :: this
    class(gradient_class),             intent(inout) :: g
    class(vector_class),               intent(in)    :: x
    class(inner_product_class),        intent(inout) :: ip

    type(trivial_inner_product_type)   :: ip_euclid

    integer :: n

    !--- define the cost function with respect to the Euclidean inner product
    ip_euclid%scalp(:)          = 1.0
    ip_euclid%one_over_scalp(:) = 1.0

    ip_euclid%name = 'euclid'

    select type (g)
    type is (trivial_gradient_type)
      select type (x)
      type is (trivial_vector_type)
        select type (ip)
        type is (trivial_inner_product_type)
          !--- change to the Euclidean inner product
          call g%change_inner_product_to(ip_euclid)
          
          !--- our cost function is J=(1/2)*x'*A*x + b'*x where A=tridiag(0.2,1,0.2)
          !--- and b=0.5
          
          n = size(x%elements)
          
          g%elements(:)     = x%elements(:) + 0.5
          g%elements(2:n)   = g%elements(2:n)   + 0.2*x%elements(1:n-1)
          g%elements(1:n-1) = g%elements(1:n-1) + 0.2*x%elements(2:n)
          
          !--- define the output gradient with respect to the requested inner product
          
          call g%change_inner_product_to(ip)
        end select
      end select
    end select
  end subroutine my_eval
end module concrete_simulator
